from __future__ import unicode_literals

from datetime import timedelta

from six import StringIO

from django.contrib.auth.models import User
from django.core.management import call_command
from django.utils import timezone

from germanium.test_cases.default import GermaniumTestCase
from germanium.tools import assert_equal, assert_false

from is_core.auth_token.models import Token
from is_core.config import settings


class CleanTokensCommandTestCase(GermaniumTestCase):

    def test_clean_tokens_remove_only_old_tokens(self):
        user = User.objects._create_user('test', 'test@test.cz', 'test', is_staff=False, is_superuser=False)
        expired_tokens = [Token.objects.create(user=user, ip='127.0.0.1') for _ in range(10)]
        not_expired_tokens = [Token.objects.create(user=user, ip='127.0.0.1') for _ in
                              range(settings.AUTH_COUNT_USER_PRESERVED_TOKENS - 5)]
        Token.objects.filter(pk__in=[token.pk for token in expired_tokens]).update(
            last_access=timezone.now() - timedelta(seconds=settings.AUTH_MAX_TOKEN_AGE))
        call_command('clean_tokens', stdout=StringIO(), stderr=StringIO())
        assert_equal(Token.objects.filter(pk__in=[token.pk for token in not_expired_tokens]).count(),
                     settings.AUTH_COUNT_USER_PRESERVED_TOKENS - 5)
        assert_equal(Token.objects.filter(pk__in=[token.pk for token in expired_tokens]).count(), 5)



from django.utils.translation import ugettext

from pyston.response import RESTErrorResponse, RESTNoContentResponse
from pyston.exception import RESTException

from is_core.rest.resource import RESTResource
from is_core.auth_token.utils import login, logout
from is_core.auth_token.forms import TokenAuthenticationForm


class AuthResource(RESTResource):
    login_required = False
    allowed_methods = ('post', 'delete')
    form_class = TokenAuthenticationForm

    def _sucessful_login(self, request):
        pass

    def _login(self, user, expiration, form):
        login(self.request, user, expiration)

    def _unsucessful_login(self, request):
        pass

    def get_form_kwargs(self):
        return {'data': self.get_dict_data()}

    def get_form_class(self):
        return self.form_class

    def post(self):
        if not self.request.data:
            raise RESTException(ugettext('Missing data'))
        form = self.get_form_class()(**self.get_form_kwargs())

        errors = form.is_invalid()
        if errors:
            self._unsucessful_login(self.request)
            return RESTErrorResponse(errors)

        self._sucessful_login(self.request)
        self._login(form.get_user(), not form.is_permanent(), form)
        return {'token': self.request.token.key, 'user': form.get_user()}

    def delete(self):
        if self.request.user.is_authenticated:
            logout(self.request)
        return RESTNoContentResponse()

    @classmethod
    def __init_core__(cls, core, pattern):
        cls.core = core
        cls.pattern = pattern

    def has_delete_permission(self, **kwargs):
        return (
            self.request.user.is_authenticated and super(AuthResource, self).has_delete_permission(**kwargs)
        )



from django import forms
from django.utils.translation import ugettext_lazy as _

from is_core.forms import SmartForm

from class_based_auth_views.forms import AuthenticationMixin, AuthenticationCleanMixin


class TokenAuthenticationMixin(AuthenticationMixin):

    def __init__(self, *args, **kwargs):
        super(TokenAuthenticationMixin, self).__init__(*args, **kwargs)
        self.fields['permanent'] = forms.BooleanField(label=_('Remember user'), required=False)

    def is_permanent(self):
        return self.cleaned_data.get('permanent', False)


class TokenAuthenticationForm(TokenAuthenticationMixin, AuthenticationCleanMixin, SmartForm):
    pass


from django.middleware.csrf import rotate_token
from django.contrib.auth import load_backend
from django.contrib.auth.signals import user_logged_in, user_logged_out
from django.contrib.auth.models import AnonymousUser
from django.conf import settings as django_settings

from ipware.ip import get_ip

from is_core.config import settings
from is_core.auth_token.models import Token, AnonymousToken
from is_core.utils import header_name_to_django
from is_core.utils.compatibility import _get_backends


def login(request, user, expiration=True, auth_slug=None, related_objs=None, backend=None):
    """
    Persist token into database. Token is stored inside cookie therefore is not necessary
    reauthenticate user for every request.
    """
    related_objs = related_objs if related_objs is not None else ()

    if user is None:
        user = request.user

    try:
        backend = backend or user.backend
    except AttributeError:
        backends = _get_backends(return_tuples=True)
        if len(backends) == 1:
            _, backend = backends[0]
        else:
            raise ValueError(
                'You have multiple authentication backends configured and '
                'therefore must provide the `backend` argument or set the '
                '`backend` attribute on the user.'
            )

    token = Token.objects.create(user=user, user_agent=request.META.get('HTTP_USER_AGENT', '')[:256],
                                 expiration=expiration, auth_slug=auth_slug, ip=get_ip(request),
                                 backend=backend)

    for related_obj in related_objs:
        token.related_objects.create(content_object=related_obj)
    if hasattr(request, 'user'):
        request.user = user
    request.token = token
    rotate_token(request)
    user_logged_in.send(sender=user.__class__, request=request, user=user)


def logout(request):
    """
    Set current token to inactive.
    """
    # Dispatch the signal before the user is logged out so the receivers have a
    # chance to find out *who* logged out.
    user = getattr(request, 'user', None)
    if hasattr(user, 'is_authenticated') and not user.is_authenticated:
        user = None
    user_logged_out.send(sender=user.__class__, request=request, user=user)

    if hasattr(request, 'token') and request.token.is_active:
        if request.token.active_takeover:
            active_takeover = request.token.active_takeover
            active_takeover.is_active = False
            active_takeover.save()

            if hasattr(request, 'user'):
                request.user = request.token.user
        else:
            token = request.token
            token.is_active = False
            token.save()

            if hasattr(request, 'user'):
                request.user = AnonymousUser()


def get_token(request):
    """
    Returns the token model instance associated with the given request token key.
    If no user is retrieved AnonymousToken is returned.
    """

    auth_token = (request.META.get(header_name_to_django(settings.AUTH_HEADER_NAME)) or
                  request.COOKIES.get(settings.AUTH_COOKIE_NAME))

    try:
        token = Token.objects.get(key=auth_token, is_active=True)
        if not token.is_expired:
            if auth_token == request.META.get(header_name_to_django(settings.AUTH_HEADER_NAME)):
                token.is_from_header = True
            return token
    except Token.DoesNotExist:
        pass
    return AnonymousToken()


def dont_enforce_csrf_checks(request):
    return (getattr(request, '_dont_enforce_csrf_checks', False) or
                request.META.get(header_name_to_django(settings.AUTH_HEADER_NAME)))


def get_user_from_token(token):
    if token:
        backend_path = token.backend
        if backend_path in django_settings.AUTHENTICATION_BACKENDS:
            active_takeover_id = token.active_takeover.user.pk if token.active_takeover else None
            user_id = token.user.pk
            backend = load_backend(backend_path)
            return backend.get_user(active_takeover_id) or backend.get_user(user_id) or AnonymousUser()
    return AnonymousUser()


def get_user(request):
    """
    Returns the user model instance associated with the given request token.
    If no user is retrieved an instance of `AnonymousUser` is returned.
    """
    return get_user_from_token(getattr(request, 'token'))


def takeover(request, user):
    if request.user == user:
        return False
    else:
        request.token.user_takeovers.update(is_active=False)
        request.token.user_takeovers.create(user=user, is_active=True)
        return True



from class_based_auth_views.views import LoginView

from django.views.generic.base import TemplateView
from django.contrib import auth


class LoginView(LoginView):
    pass


class LogoutView(TemplateView):
    template_name = "registration/logout.html"

    def get(self, *args, **kwargs):
        if self.request.user.is_authenticated:
            auth.logout(self.request)
        return super(LogoutView, self).get(*args, **kwargs)



from .version import *


# Apply patch only if django is installed
try:
    from django.core.exceptions import ImproperlyConfigured
    try:
        from django.db import models
        from .filters.patch import *
        from .models.patch import *
        from .forms.patch import *
    except ImproperlyConfigured:
        pass
except ImportError as ex:
    pass



from django.db import models, migrations


class Migration(migrations.Migration):

    dependencies = [
        ('contenttypes', '0001_initial'),
        ('auth_token', '0001_initial'),
    ]

    operations = [
        migrations.CreateModel(
            name='TokenRelatedObject',
            fields=[
                ('id', models.AutoField(verbose_name='ID', serialize=False, auto_created=True, primary_key=True)),
                ('object_id', models.TextField()),
                ('content_type', models.ForeignKey(to='contenttypes.ContentType',
                                           on_delete=models.CASCADE)),
                ('token', models.ForeignKey(related_name='related_objects', to='auth_token.Token',
                                           on_delete=models.CASCADE)),
            ],
        ),
    ]



from django.views.generic.base import TemplateView

from is_core.generic_views import DefaultViewMixin
from is_core.patterns import patterns, RESTPattern


class RESTDocumentationView(DefaultViewMixin, TemplateView):
    template_name = 'generic_views/rest_documentation.html'

    def get_context_data(self, **kwargs):
        context_data = super(RESTDocumentationView, self).get_context_data(**kwargs)
        rest_patterns = {}
        for pattern_name, pattern in patterns.items():
            if isinstance(pattern, RESTPattern):
                rest_patterns[pattern_name] = pattern
        context_data['pattern_list'] = rest_patterns
        return context_data




from django.urls.resolvers import get_resolver

from germanium.test_cases.client import ClientTestCase
from germanium.annotations import login_all, data_provider
from germanium.tools.http import assert_http_ok

from is_core.tests.data_generator_test_case import DataGeneratorTestCase
from is_core.utils.compatibility import reverse


class ModelUITestCaseMiddleware(DataGeneratorTestCase):

    add_disabled_views = ()
    edit_disabled_views = ()
    list_disabled_views = ()
    ignore_warnings = False

    @classmethod
    def setUpClass(cls):
        super(ModelUITestCaseMiddleware, cls).setUpClass()
        cls.ui_main_views = cls.set_up_main_views()

    @classmethod
    def set_up_main_views(cls):
        from is_core.site import registered_model_cores
        from is_core.main import UIRESTModelISCore

        # Must be here, because hanlers is not registered
        get_resolver().url_patterns

        ui_main_views = []
        for main_view in [model_view for model_view in registered_model_cores.values() if isinstance(model_view,
                                                                                                     UIRESTModelISCore)]:
            model = main_view.model
            if cls.get_model_label(model) in cls.factories:
                ui_main_views.append((main_view, model))
            elif not cls.ignore_warnings:
                cls.logger.warning('Model {} has not created factory class'.format(model))

        return ui_main_views

    def list_url(self, site_name, menu_groups):
        return reverse('{}:list-{}'.format(site_name, '-'.join(menu_groups)))

    def add_url(self, site_name, menu_groups):
        return reverse('{}:add-{}'.format(site_name, '-'.join(menu_groups)))

    def edit_url(self, site_name, menu_groups, obj):
        return reverse('{}:edit-{}'.format(site_name, '-'.join(menu_groups)), args=(obj.pk,))

    def view_name(self, model_view):
        return '{}-{}'.format(model_view.site_name, '-'.join(model_view.get_menu_groups()))


@login_all
class TestSiteAvailability(ModelUITestCaseMiddleware, ClientTestCase):

    def get_ui_main_views(self):
        return self.ui_main_views

    @data_provider(get_ui_main_views)
    def test_should_return_right_list_page_for_all_model_views(self, model_view, model):

        if 'list' in model_view.ui_patterns:
            url = self.list_url(model_view.site_name, model_view.get_menu_groups())
            if model_view.has_ui_read_permission(self.get_request_with_user(self.r_factory.get(url))):
                assert_http_ok(self.get(url), '{} should return 200'.format(url))

    @data_provider(get_ui_main_views)
    def test_should_return_right_add_page_for_all_model_view(self, model_view, model):

        if 'add' in model_view.ui_patterns:
            url = self.add_url(model_view.site_name, model_view.get_menu_groups())
            if model_view.has_ui_create_permission(self.get_request_with_user(self.r_factory.get(url))):
                assert_http_ok(self.get(url), '{} should return 200'.format(url))

    @data_provider(get_ui_main_views)
    def test_should_return_right_edit_page_for_all_model_view(self, model_view, model):

        if 'edit' in model_view.ui_patterns:
            inst = self.new_instance(model)

            url = self.edit_url(model_view.site_name, model_view.get_menu_groups(), inst)
            request = self.get_request_with_user(self.r_factory.get(url))
            if model_view.has_ui_read_permission(request, inst) or model_view.has_ui_update_permission(request, inst.pk):
                assert_http_ok(self.get(url), '{} should return 200'.format(url))



import os, binascii

from datetime import timedelta

from django.conf import settings as django_settings
from django.db import models
from django.contrib.auth.models import AnonymousUser
from django.contrib.contenttypes.models import ContentType
from django.utils import timezone
from django.utils.encoding import force_text

try:
    from django.contrib.contenttypes.fields import GenericForeignKey
except ImportError:  # Django < 1.9 pragma: no cover
    from django.contrib.contenttypes.generic import GenericForeignKey

from is_core.config import settings


# Prior to Django 1.5, the AUTH_USER_MODEL setting does not exist.
AUTH_USER_MODEL = getattr(django_settings, 'AUTH_USER_MODEL', 'auth.User')


class Token(models.Model):
    """
    The default authorization token model.
    """

    key = models.CharField(max_length=40, primary_key=True, null=False, blank=False)
    user = models.ForeignKey(AUTH_USER_MODEL, related_name='auth_token', null=False, blank=False,
                             on_delete=models.CASCADE)
    created_at = models.DateTimeField(auto_now_add=True, null=False, blank=False)
    last_access = models.DateTimeField(auto_now=True, null=False, blank=False)
    is_active = models.BooleanField(default=True)
    # TODO: It is possiple use https://github.com/selwin/django-user_agents/tree/master/django_user_agents or
    # https://github.com/selwin/python-user-agents for parse
    # Limited size to 256
    user_agent = models.CharField(max_length=256, null=True, blank=True)
    expiration = models.BooleanField(null=False, default=True)
    ip = models.GenericIPAddressField(null=False, blank=False)
    auth_slug = models.SlugField(null=True, blank=True)
    backend = models.CharField(max_length=255, null=False, blank=False)

    is_from_header = False

    @property
    def active_takeover(self):
        return self.user_takeovers.filter(is_active=True).last()

    def save(self, *args, **kwargs):
        if not self.key:
            self.key = self.generate_key()
        return super(Token, self).save(*args, **kwargs)

    def generate_key(self):
        """
        Random id generating
        """
        return force_text(binascii.hexlify(os.urandom(20)))

    def _get_token_age(self):
        return self.expiration and settings.AUTH_DEFAULT_TOKEN_AGE or settings.AUTH_MAX_TOKEN_AGE

    @property
    def is_expired(self):
        return self.last_access + timedelta(seconds=self._get_token_age()) < timezone.now()

    @property
    def time_to_expiration(self):
        return (self.last_access + timedelta(seconds=self._get_token_age())) - timezone.now()

    @property
    def str_time_to_expiration(self):
        return str(self.time_to_expiration) if self.time_to_expiration.total_seconds() > 0 else '00:00:00'

    def __str__(self):
        return self.key


class TokenRelatedObject(models.Model):
    """
    Generic relation to objects related with authorization token
    """

    token = models.ForeignKey(Token, related_name='related_objects', on_delete=models.CASCADE)
    content_type = models.ForeignKey(ContentType, on_delete=models.CASCADE)
    object_id = models.TextField()
    content_object = GenericForeignKey('content_type', 'object_id')


class UserTokenTakeover(models.Model):
    """
    The model allows to change user without token change
    """

    token = models.ForeignKey(Token, related_name='user_takeovers', on_delete=models.CASCADE)
    user = models.ForeignKey(AUTH_USER_MODEL, related_name='user_token_takeovers', null=False, blank=False,
                             on_delete=models.CASCADE)
    is_active = models.BooleanField()


class AnonymousToken:
    key = None
    user = AnonymousUser
    creted_at = None
    is_active = False
    user_agent = None
    is_expired = True
    is_from_header = False
    active_takeover = None
    backend = None

    def save(self):
        raise NotImplementedError

    def delete(self):
        raise NotImplementedError



import re

from django.views.generic.base import TemplateView, View
from django.utils.translation import ugettext_lazy as _
from django.conf import settings

from block_snippets.views import JSONSnippetTemplateResponseMixin

from is_core.menu import LinkMenuItem
from is_core.exceptions import HTTPForbiddenResponseException
from is_core.generic_views.exceptions import GenericViewException
from django.http.response import Http404


class PermissionsViewMixin:

    def dispatch(self, request, *args, **kwargs):
        rm = request.method.lower()
        if rm in self.http_method_names and hasattr(self, rm):
            self._check_permission(rm)
            handler = getattr(self, rm)
        else:
            handler = self.http_method_not_allowed
        return handler(request, *args, **kwargs)

    def __getattr__(self, name):
        for regex, method in (
                (r'_check_(\w+)_permission', self._check_permission),
                (r'can_call_(\w+)', self._check_call)):
            m = re.match(regex, name)
            if m:
                def _call(*args, **kwargs):
                    return method(m.group(1), *args, **kwargs)
                return _call
        raise AttributeError("%r object has no attribute %r" % (self.__class__, name))

    def _check_call(self, name, *args, **kwargs):
        if not hasattr(self, 'has_%s_permission' % name):
            if settings.DEBUG:
                raise NotImplementedError('Please implement method has_%s_permission to %s' % (name, self.__class__))
            else:
                return False

        try:
            return getattr(self, 'has_%s_permission' % name)(*args, **kwargs)
        except Http404:
            return False

    def _check_permission(self, name, *args, **kwargs):
        if not hasattr(self, 'has_%s_permission' % name):
            if settings.DEBUG:
                raise NotImplementedError('Please implement method has_%s_permission to %s' % (name, self.__class__))
            else:
                raise HTTPForbiddenResponseException

        if not getattr(self, 'has_%s_permission' % name)(*args, **kwargs):
            raise HTTPForbiddenResponseException

    def has_options_permission(self, **kwargs):
        return True


class DefaultViewMixin(PermissionsViewMixin, JSONSnippetTemplateResponseMixin):

    site_name = None
    menu_groups = None
    login_required = True
    allowed_snippets = ('content',)
    view_name = None
    add_current_to_breadcrumbs = True
    kwargs = {}
    args = {}
    title = None
    page_title = None

    def __init__(self):
        super(DefaultViewMixin, self).__init__()

    def get_title(self):
        return self.title

    def get_page_title(self):
        return self.page_title or self.get_title()

    def get_context_data(self, **kwargs):
        context_data = super(DefaultViewMixin, self).get_context_data(**kwargs)
        extra_context_data = {
            'site_name': self.site_name,
            'active_menu_groups': self.menu_groups,
            'title': self.get_title(),
            'page_title': self.get_page_title(),
            'view_name': self.view_name,
            'bread_crumbs_menu_items': self.bread_crumbs_menu_items(),
        }
        extra_context_data.update(context_data)
        return extra_context_data

    def bread_crumbs_menu_items(self):
        if self.add_current_to_breadcrumbs:
            return [LinkMenuItem(self.get_title(), self.request.path, active=True)]
        return []


class DefaultCoreViewMixin(DefaultViewMixin):

    core = None
    view_name = None
    title = None

    def __init__(self):
        super(DefaultCoreViewMixin, self).__init__()
        self.site_name = self.core.site_name
        self.menu_groups = self.core.get_menu_groups()

    @classmethod
    def __init_core__(cls, core, pattern):
        cls.core = core
        cls.pattern = pattern

    def dispatch(self, request, *args, **kwargs):
        self.core.init_ui_request(request)
        return super(DefaultCoreViewMixin, self).dispatch(request, *args, **kwargs)

    def get_title(self):
        return self.title or self.model._meta.verbose_name

    @property
    def view_name(self):
        return '%s-%s' % (self.view_type, '-'.join(self.menu_groups))

    def get_permissions(self):
        return {
            'read': self.core.has_ui_read_permission,
            'create': self.core.has_ui_create_permission,
            'update': self.core.has_ui_update_permission,
            'delete': self.core.has_ui_delete_permission,
        }

    def get_context_data(self, **kwargs):
        context_data = super(DefaultCoreViewMixin, self).get_context_data(**kwargs)
        extra_context_data = {
            'permissions': self.get_permissions(),
        }
        extra_context_data.update(context_data)
        return extra_context_data


class DefaultModelCoreViewMixin(DefaultCoreViewMixin):

    model = None

    def __init__(self):
        super(DefaultModelCoreViewMixin, self).__init__()

    @classmethod
    def __init_core__(cls, core, pattern):
        super(DefaultModelCoreViewMixin, cls).__init_core__(core, pattern)
        cls.model = cls.model or core.model

    def get_context_data(self, **kwargs):
        context_data = super(DefaultModelCoreViewMixin, self).get_context_data(**kwargs)
        extra_context_data = {
            'core': self.core
        }
        extra_context_data.update(context_data)
        return extra_context_data


class HomeView(DefaultCoreViewMixin, TemplateView):
    template_name = 'home.html'
    view_name = 'home'

    def get_title(self):
        return _('Home')

    def has_get_permission(self, **kwargs):
        return True



from distutils.version import StrictVersion

import django
from django.forms.forms import BoundField as OriginalBoundField


try:
    from django.core.urlresolvers import reverse, resolve, NoReverseMatch
except ImportError:
    from django.urls import reverse, resolve, NoReverseMatch


try:
    from django.apps import apps
    get_model = apps.get_model
except ImportError:
    from django.db.models.loading import get_model


try:
    from django.template import Library
except ImportError:
    from django.template.base import Library


def admin_display_for_value(value):
    if StrictVersion(django.get_version()) >= StrictVersion('1.9'):
        from django.contrib.admin.utils import display_for_value

        return display_for_value(value, '-')
    else:
        try:
            from django.contrib.admin.utils import display_for_value

            return display_for_value(value)
        except ImportError:
            from django.contrib.admin.util import display_for_value

            return display_for_value(value)


def urls_wrapper(*urls):
    if StrictVersion(django.get_version()) >= StrictVersion('1.9'):
        return list(urls)
    else:
        from django.conf.urls import patterns

        return patterns('', *urls)


def get_model_name(model):
    return str(model._meta.model_name)


class BoundField(OriginalBoundField):

    def build_widget_attrs(self, attrs, widget=None):
        if StrictVersion(django.get_version()) >= StrictVersion('1.10'):
            return super(BoundField, self).build_widget_attrs(attrs, widget)
        else:
            return attrs


try:
    from django.contrib.auth import _get_backends
    get_model = apps.get_model
except ImportError:
    from django.contrib.auth import get_backends
    _get_backends = get_backends


class CompatibilitySelectMixin:

    if StrictVersion(django.get_version()) >= StrictVersion('1.10'):
        def render(self, name, value, attrs=None, choices=()):
            return super(CompatibilitySelectMixin, self).render(name, value, attrs=attrs)

        def render_options(self, selected_choices):
            if hasattr(self, 'render_options_compatible'):
                return self.render_options_compatible((), selected_choices)
            else:
                return super(CompatibilitySelectMixin, self).render_options(selected_choices)
    else:
        def render(self, name, value, attrs=None, choices=()):
            return super(CompatibilitySelectMixin, self).render(name, value, attrs=attrs, choices=choices)

        def render_options(self, choices, selected_choices):
            if hasattr(self, 'render_options_compatible'):
                return self.render_options_compatible(choices, selected_choices)
            else:
                return super(CompatibilitySelectMixin, self).render_options(choices, selected_choices)


class CompatibilityWidgetMixin:

    def build_attrs(self, base_attrs, extra_attrs=None, **kwargs):
        """
        Helper function for building an attribute dictionary.
        This is combination of the same method from Django<=1.10 and Django1.11+
        """

        attrs = dict(base_attrs, **kwargs)
        if extra_attrs:
            attrs.update(extra_attrs)
        return attrs


class MiddlewareMixin:

    def __init__(self, get_response=None):
        self.get_response = get_response
        super().__init__()

    def __call__(self, request):
        response = None
        if hasattr(self, 'process_request'):
            response = self.process_request(request)
        response = response or self.get_response(request)
        if hasattr(self, 'process_response'):
            response = self.process_response(request, response)
        return response



from datetime import timedelta

from django.utils import timezone
from django.core.management.base import BaseCommand

from is_core.config import settings
from is_core.auth_token.models import Token, AUTH_USER_MODEL
from is_core.utils.compatibility import get_model


class Command(BaseCommand):

    def handle(self, **options):
        for user in get_model(*AUTH_USER_MODEL.split('.', 1)).objects.all():
            user_last_preserved_token_pks = Token.objects.filter(
                user=user
            ).order_by('-created_at')[:settings.AUTH_COUNT_USER_PRESERVED_TOKENS].values_list('pk', flat=True)
            removing_tokens_qs = Token.objects.filter(
                last_access__lt=timezone.now() - timedelta(seconds=settings.AUTH_MAX_TOKEN_AGE),
                user=user
            ).exclude(pk__in=user_last_preserved_token_pks)
            self.stdout.write('Removing {} tokens of user {}'.format(removing_tokens_qs.count(), user))
            removing_tokens_qs.delete()



from is_core.auth_token.default.resource import AuthResource as DefaultAuthResource
from is_core.auth_token.auth_security import LOGIN_THROTTLING_VALIDATORS

from security.models import InputLoggedRequest
from security.decorators import hide_request_body_all, throttling


@hide_request_body_all
class AuthResource(DefaultAuthResource):

    @throttling(*LOGIN_THROTTLING_VALIDATORS)
    def post(self):
        return super(AuthResource, self).post()

    def _sucessful_login(self, request):
        if getattr(self.request, '_logged_request', False):
            self.request._logged_request.type = InputLoggedRequest.SUCCESSFUL_LOGIN_REQUEST

    def _unsucessful_login(self, request):
        if getattr(self.request, '_logged_request', False):
            self.request._logged_request.type = InputLoggedRequest.UNSUCCESSFUL_LOGIN_REQUEST



from django.conf import settings
from django.db import migrations, models
import django.db.models.deletion


class Migration(migrations.Migration):

    dependencies = [
        migrations.swappable_dependency(settings.AUTH_USER_MODEL),
        ('auth_token', '0002_tokenrelatedobject'),
    ]

    operations = [
        migrations.CreateModel(
            name='UserTokenTakeover',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('is_active', models.BooleanField()),
                ('token', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE,
                                            related_name='user_takeovers', to='auth_token.Token')),
                ('user', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE,
                                           related_name='user_token_takeovers', to=settings.AUTH_USER_MODEL)),
            ],
        ),
    ]



from functools import wraps

from django.utils.decorators import available_attrs


def auth_token_renewal_exempt(view_func):
    def wrapped_view(*args, **kwargs):
        return view_func(*args, **kwargs)
    wrapped_view.auth_token_renewal_exempt = True
    return wraps(view_func, assigned=available_attrs(view_func))(wrapped_view)



from functools import wraps

from django.contrib.auth.decorators import login_required
from django.http.response import HttpResponseForbidden
from django.utils.functional import SimpleLazyObject
from django.utils.decorators import available_attrs

from chamber.shortcuts import get_object_or_none

from is_core import config
from is_core.exceptions import HTTPUnauthorizedResponseException


def rest_login_required(rest_func):

    @wraps(rest_func, assigned=available_attrs(rest_func))
    def _rest_login_required(request, *args, **kwargs):
        if not hasattr(request, 'user') or not request.user or not request.user.is_authenticated:
            raise HTTPUnauthorizedResponseException
        return rest_func(request, *args, **kwargs)

    return _rest_login_required



from security.models import InputLoggedRequest
from security.decorators import hide_request_body_all, throttling_all

from is_core.auth_token.default.views import TokenLoginView as DefaultTokenLoginView
from is_core.auth_token.auth_security import LOGIN_THROTTLING_VALIDATORS


@throttling_all(*LOGIN_THROTTLING_VALIDATORS)
@hide_request_body_all
class TokenLoginView(DefaultTokenLoginView):

    def form_valid(self, form):
        if getattr(self.request, '_logged_request', False):
            self.request._logged_request.type = InputLoggedRequest.SUCCESSFUL_LOGIN_REQUEST
        return super(TokenLoginView, self).form_valid(form)

    def form_invalid(self, form):
        if getattr(self.request, '_logged_request', False):
            self.request._logged_request.type = InputLoggedRequest.UNSUCCESSFUL_LOGIN_REQUEST
        return super(TokenLoginView, self).form_invalid(form)



def merge(origin, *args):
    """
    Merges given dictionaries, `origin` will not be changed.
    TODO: Remove this function after `django-chamber` upgrade
    """
    copy = origin.copy()
    for dictionary in args:
        copy.update(dictionary)
    return copy



from django.http.response import HttpResponseRedirect
from django.views.generic.base import RedirectView

from is_core.config import settings
from is_core.generic_views import DefaultCoreViewMixin
from is_core.generic_views.mixins import GetCoreObjViewMixin
from is_core.generic_views.auth_views import LogoutView, LoginView
from is_core.auth_token.utils import login, logout, takeover
from is_core.auth_token.forms import TokenAuthenticationForm


class TokenLoginView(LoginView):

    form_class = TokenAuthenticationForm

    def _login(self, user, expiration, form):
        login(self.request, user, expiration)

    def dispatch(self, request, *args, **kwargs):
        if request.user and request.user.is_authenticated:
            return HttpResponseRedirect(self.get_success_url())
        return super(TokenLoginView, self).dispatch(request, *args, **kwargs)

    def form_valid(self, form):
        """
        The user has provided valid credentials (this was checked in AuthenticationForm.is_valid()). So now we
        can check the test cookie stuff and log him in.
        """
        self.check_and_delete_test_cookie()
        self._login(form.get_user(), not form.is_permanent(), form)
        return super(TokenLoginView, self).form_valid(form)


class TokenLogoutView(LogoutView):

    def get(self, *args, **kwargs):
        if self.request.user.is_authenticated:
            logout(self.request)
        return super(LogoutView, self).get(*args, **kwargs)


class UserTakeover(GetCoreObjViewMixin, DefaultCoreViewMixin, RedirectView):

    def get_redirect_url(self, *args, **kwargs):
        return settings.AUTH_TAKEOVER_REDIRECT_URL

    def get(self, request, *args, **kwargs):
        user = self.get_obj()
        takeover(self.request, user)
        return super(UserTakeover, self).get(request, *args, **kwargs)

    def has_get_permission(self, *args, **kwargs):
        return self.request.user.is_superuser



import time

from django.utils.encoding import force_text
from django.utils.functional import SimpleLazyObject
from django.utils.http import cookie_date

from is_core.auth_token import utils
from is_core.config import settings
from is_core.utils import header_name_to_django
from is_core.utils.compatibility import MiddlewareMixin


def get_user(request):
    if not hasattr(request, '_cached_user'):
        request._cached_user = utils.get_user(request)
    return request._cached_user


class TokenAuthenticationMiddlewares(MiddlewareMixin):

    def process_request(self, request):
        """
        Lazy set user and token
        """
        request.token = utils.get_token(request)
        request.user = SimpleLazyObject(lambda: get_user(request))
        request._dont_enforce_csrf_checks = utils.dont_enforce_csrf_checks(request)

    def _update_token_and_cookie(self, request, response, max_age, expires):
        request.token.save()
        response.set_cookie(settings.AUTH_COOKIE_NAME, force_text(request.token.key), max_age=max_age,
                            expires=expires, httponly=settings.AUTH_COOKIE_HTTPONLY,
                            secure=settings.AUTH_COOKIE_SECURE, domain=settings.AUTH_COOKIE_DOMAIN)
        return response

    def _set_auth_expiration_header(self, request, response):
        response[settings.AUTH_TOKEN_EXPIRATION_HEADER] = request.token.str_time_to_expiration

    def _is_token_renewal_exempt(self, request):
        return (getattr(request, 'auth_token_renewal_exempt', False) or
                header_name_to_django(settings.AUTH_TOKEN_RENEWAL_EXEMPT_HEADER) in request.META)

    def process_response(self, request, response):
        """
        Set cookie with token key if user is authenticated
        """
        # Save the session data and refresh the client cookie.
        # Skip session save for 500 responses, refs #3881.
        if response.status_code != 500 and hasattr(request, 'token') and request.token.is_active:
            if request.token.expiration:
                # The user did not choose to be permanently signed. Hence, the authentication cookie that holds the
                # token value is set to expire when the browser is closed
                max_age = None
                expires = None
            else:
                max_age = settings.AUTH_COOKIE_AGE
                expires = cookie_date(time.time() + max_age)

            if not self._is_token_renewal_exempt(request):
                self._update_token_and_cookie(request, response, max_age, expires)

            self._set_auth_expiration_header(request, response)

        return response

    def process_view(self, request, view_func, view_args, view_kwargs):
        request.auth_token_renewal_exempt = getattr(view_func, 'auth_token_renewal_exempt', False)



from __future__ import unicode_literals

from django.db import models
from django.utils.translation import ugettext_lazy as _
from django.conf import settings


AUTH_USER_MODEL = getattr(settings, 'AUTH_USER_MODEL', 'auth.User')


class Issue(models.Model):

    name = models.CharField(verbose_name=_('Name'), max_length=100, null=False, blank=False)
    watched_by = models.ManyToManyField(AUTH_USER_MODEL, verbose_name=_('Watched by'), blank=True,
                                        related_name='watching_issues')
    created_by = models.ForeignKey(AUTH_USER_MODEL, verbose_name=_('Created by'), null=False, blank=False,
                                   related_name='created_issues', on_delete=models.CASCADE)

    solver = models.OneToOneField(AUTH_USER_MODEL, verbose_name=_('Solver'), null=True, blank=True,
                                  related_name='solving_issue', on_delete=models.CASCADE)
    leader = models.OneToOneField(AUTH_USER_MODEL, verbose_name=_('Leader'), null=False, blank=False,
                                  related_name='leading_issue', on_delete=models.CASCADE)

    def watched_by_string(self):
        return ', '.join(self.watched_by.all())
    watched_by_string.order_by = 'watched_by'

    def __unicode__(self):
        return 'issue: %s' % self.name



from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('auth_token', '0003_usertokentakeover'),
    ]

    operations = [
        migrations.AddField(
            model_name='token',
            name='backend',
            field=models.CharField(default='django.contrib.auth.backends.ModelBackend', max_length=255),
            preserve_default=False,
        ),
    ]



from django.template.base import Node, TemplateSyntaxError, NodeList
from django.utils.functional import SimpleLazyObject

from is_core.utils.compatibility import Library


register = Library()


class Permissions:
    permissions_validators = {}

    def register_permission_validator(self, name, validator):
        self.permissions_validators[name] = validator

permissions = Permissions()


def get_obj(Model, pk):
    return Model.objects.get(pk=pk)


# TODO: there is possibility cache permissions
class PermissionNode(Node):
    def __init__(self, perm_name, vals, nodelist_true, nodelist_false):
        self.perm_name = perm_name
        self.nodelist_true = nodelist_true
        self.nodelist_false = nodelist_false
        self.vals = vals

    def render(self, context):
        perm_name = self.perm_name.resolve(context, True)
        request = context.get('request')
        args = []
        for val in self.vals:
            args.append(val.resolve(context))

        view_permissions = context.get('permissions', {})

        if perm_name in view_permissions:
            perm_fun_or_bool = view_permissions.get(perm_name)
            if ((isinstance(perm_fun_or_bool, bool) and perm_fun_or_bool) or
                    (hasattr(perm_fun_or_bool, '__call__') and perm_fun_or_bool(request, *args))):
                return self.nodelist_true.render(context)

        if perm_name in permissions.permissions_validators:
            request = context.get('request')
            if permissions.permissions_validators.get(perm_name)(request, *args):
                return self.nodelist_true.render(context)
        return self.nodelist_false.render(context)

    def validator_kwargs(self, request, validator):
        if 'pk' in request.kwargs:
            Model = getattr(validator.im_self, 'model')
            if Model:
                return {'obj': SimpleLazyObject(lambda: get_obj(Model, request.kwargs['pk']))}
        return {}


@register.tag
def has_permission(parser, token):
    bits = list(token.split_contents())
    if len(bits) < 2:
        raise TemplateSyntaxError("%r takes minimal one argument" % bits[0])
    end_tag = 'end' + bits[0]

    vals = []
    for bit in bits[2:]:
        vals.append(parser.compile_filter(bit))

    nodelist_true = parser.parse(('else', end_tag))
    token = parser.next_token()
    if token.contents == 'else':
        nodelist_false = parser.parse((end_tag,))
        parser.delete_first_token()
    else:
        nodelist_false = NodeList()
    return PermissionNode(parser.compile_filter(bits[1]), vals, nodelist_true, nodelist_false)



from django.db import models, migrations
from django.conf import settings


class Migration(migrations.Migration):

    dependencies = [
        migrations.swappable_dependency(settings.AUTH_USER_MODEL),
    ]

    operations = [
        migrations.CreateModel(
            name='Token',
            fields=[
                ('key', models.CharField(max_length=40, primary_key=True, serialize=False)),
                ('created_at', models.DateTimeField(auto_now_add=True)),
                ('last_access', models.DateTimeField(auto_now=True)),
                ('is_active', models.BooleanField(default=True)),
                ('user_agent', models.CharField(max_length=256, null=True, blank=True)),
                ('expiration', models.BooleanField(default=True)),
                ('ip', models.GenericIPAddressField()),
                ('auth_slug', models.SlugField(null=True, blank=True)),
                ('user', models.ForeignKey(to=settings.AUTH_USER_MODEL, related_name='auth_token',
                                           on_delete=models.CASCADE)),
            ],
        ),
    ]



from collections import OrderedDict

from django.conf.urls import url, include
from django.template.defaultfilters import lower
from django.core.exceptions import ImproperlyConfigured

from is_core.config import settings
from is_core.utils.compatibility import urls_wrapper

from .utils import str_to_class
from .loading import get_cores
from .patterns import RESTPattern
from .rest.resource import EntryPointResource


sites = {}
registered_model_cores = {}
registered_cores = []


def get_site_by_name(name):
    """
    Return site according to name (default is IS)
    """
    return sites.get(name)


def get_model_core(model):
    """
    Return core view of given model or None
    """
    model_label = lower('%s.%s' % (model._meta.app_label, model._meta.object_name))
    return registered_model_cores.get(model_label)


class ISSite:

    def __init__(self, name='IS'):
        self.name = name
        sites[name] = self
        self._registry = self._init_items()

    def _init_items(self):
        out = OrderedDict()

        for core in get_cores():
            generic_core = self.register(core(self.name, []))
            if generic_core.menu_group in out:
                raise ImproperlyConfigured('Duplicate cores with group: "%s"' % generic_core.menu_group)
            out[generic_core.menu_group] = generic_core
        return out

    def register(self, generic_core):
        if (hasattr(generic_core, 'model')):
            model_label = lower('%s.%s' % (generic_core.model._meta.app_label, generic_core.model._meta.object_name))
            registered_model_cores[model_label] = generic_core
        registered_cores.append(generic_core)
        return generic_core

    @property
    def urls(self):
        return self.get_urls(), self.name

    def _set_items_urls(self, items, urlpatterns):
        for item in items:
            urls = item.get_urls()
            if urls:
                urlpatterns += urls_wrapper(url(r'^', include(urls)))

    def get_urls(self):
        urlpatterns = urls_wrapper()
        if settings.AUTH_LOGIN_VIEW:
            urlpatterns += urls_wrapper(
                url(r'^{}$'.format(settings.LOGIN_URL[1:]), str_to_class(settings.AUTH_LOGIN_VIEW).as_view(
                    form_class=str_to_class(settings.AUTH_FORM_CLASS)), name='login'),
            )

        if settings.AUTH_LOGOUT_VIEW:
            urlpatterns += urls_wrapper(
                url(r'^{}$'.format(settings.LOGOUT_URL[1:]),
                    str_to_class(settings.AUTH_LOGOUT_VIEW).as_view(), name='logout'),
            )

        if settings.AUTH_USE_TOKENS and settings.AUTH_RESOURCE_CLASS:
            auth_resource_class = str_to_class(settings.AUTH_RESOURCE_CLASS)
            auth_resource_class.form_class = str_to_class(settings.AUTH_FORM_CLASS)
            pattern = RESTPattern('api-login', self.name, settings.LOGIN_API_URL[1:],
                                  auth_resource_class)
            urlpatterns += urls_wrapper(pattern.get_url())

        pattern = RESTPattern('api', self.name, r'api/', EntryPointResource)
        urlpatterns += urls_wrapper(pattern.get_url())

        self._set_items_urls(self._registry.values(), urlpatterns)
        return urlpatterns

site = ISSite()


def get_core(name):
    return site._registry[name]



from pyston.patch import DateField, DateTimeField, ManyToManyField, ForeignKey, ForeignObjectRel

from .default_filters import (
    DateRangeFilter, DateTimeRangeFilter, UIManyToManyFieldFilter, UIForeignKeyFilter, UIForeignObjectRelFilter
)


DateField.default_filter = DateRangeFilter
DateTimeField.default_filter = DateTimeRangeFilter
ManyToManyField.default_filter = UIManyToManyFieldFilter
ForeignKey.default_filter = UIForeignKeyFilter
ForeignObjectRel.default_filter = UIForeignObjectRelFilter



from __future__ import unicode_literals

from datetime import timedelta

from django.contrib.auth.models import User
from django.utils import timezone

from germanium.test_cases.default import GermaniumTestCase
from germanium.tools import assert_equal

from is_core.auth_token.models import Token
from is_core.config import settings


class TokenTestCase(GermaniumTestCase):

    def test_should_return_proper_string_format_for_expiration(self):
        user = User.objects._create_user('test', 'test@test.cz', 'test', is_staff=False, is_superuser=False)
        expired_token = Token.objects.create(user=user, ip='127.0.0.1')
        Token.objects.filter(pk=expired_token.pk).update(
            last_access=timezone.now() - timedelta(seconds=settings.AUTH_MAX_TOKEN_AGE))
        expired_token = Token.objects.get(pk=expired_token.pk)
        assert_equal('00:00:00', Token.objects.get(pk=expired_token.pk).str_time_to_expiration)

        non_expired_token = Token.objects.create(user=user, ip='127.0.0.1')
        assert_equal('0:59:59', non_expired_token.str_time_to_expiration.split('.')[0])



from django.conf import settings
from django.http.response import Http404
from django.utils.safestring import mark_safe
from django.utils.translation import ugettext
from django.utils.encoding import force_text

from pyston.conf import settings as pyston_settings
from pyston.forms import rest_modelform_factory
from pyston.exception import (RESTException, MimerDataException, NotAllowedException, UnsupportedMediaTypeException,
                              ResourceNotFoundException, NotAllowedMethodException, DuplicateEntryException,
                              ConflictException, DataInvalidException)
from pyston.resource import BaseResource, BaseModelResource, DefaultRESTModelResource
from pyston.response import RESTErrorResponse, RESTErrorsResponse
from pyston.utils import rfs

from chamber.shortcuts import get_object_or_none
from chamber.utils.decorators import classproperty
from chamber.utils import transaction

from is_core.config import settings
from is_core.exceptions.response import (HTTPBadRequestResponseException, HTTPUnsupportedMediaTypeResponseException,
                                         HTTPMethodNotAllowedResponseException, HTTPDuplicateResponseException,
                                         HTTPForbiddenResponseException, HTTPUnauthorizedResponseException)
from is_core.forms.models import smartmodelform_factory
from is_core.patterns import RESTPattern, patterns
from is_core.utils.immutable import merge
from is_core.utils.compatibility import NoReverseMatch


class RESTLoginMixin:

    login_required = True
    login_post_required = True
    login_patch_required = True
    login_put_required = True
    login_get_required = True
    login_delete_required = True
    login_head_required = True
    login_options_required = True

    def has_login_post_required(self):
        return self.login_required and self.login_post_required

    def has_login_put_required(self):
        return self.login_required and self.login_put_required

    def has_login_patch_required(self):
        return self.login_required and self.login_patch_required

    def has_login_get_required(self):
        return self.login_required and self.login_get_required

    def has_login_delete_required(self):
        return self.login_required and self.login_delete_required

    def has_login_head_required(self):
        return self.login_required and self.login_head_required and self.has_login_get_required()

    def has_login_options_required(self):
        return self.login_required and self.login_options_required and not self._is_cors_options_request()

    def has_get_permission(self, **kwargs):
        return ((not self.has_login_get_required() or self.request.user.is_authenticated) and
                super(RESTLoginMixin, self).has_get_permission(**kwargs))

    def has_post_permission(self, **kwargs):
        return ((not self.has_login_post_required() or self.request.user.is_authenticated) and
                super(RESTLoginMixin, self).has_post_permission(**kwargs))

    def has_put_permission(self, **kwargs):
        return ((not self.has_login_put_required() or self.request.user.is_authenticated) and
                super(RESTLoginMixin, self).has_put_permission(**kwargs))

    def has_patch_permission(self, **kwargs):
        return ((not self.has_login_patch_required() or self.request.user.is_authenticated) and
                super(RESTLoginMixin, self).has_patch_permission(**kwargs))

    def has_delete_permission(self, **kwargs):
        return ((not self.has_login_delete_required() or self.request.user.is_authenticated) and
                super(RESTLoginMixin, self).has_delete_permission(**kwargs))

    def has_options_permission(self, **kwargs):
        return ((not self.has_login_options_required() or self.request.user.is_authenticated) and
                super(RESTLoginMixin, self).has_options_permission(**kwargs))

    def dispatch(self, request, *args, **kwargs):
        if ((not hasattr(request, 'user') or not request.user or not request.user.is_authenticated) and
                getattr(self, 'has_login_{}_required'.format(request.method.lower()))()):
            raise HTTPUnauthorizedResponseException
        else:
            return super(RESTLoginMixin, self).dispatch(request, *args, **kwargs)


class RESTObjectLoginMixin(RESTLoginMixin):

    login_create_obj_required = True
    login_update_obj_required = True
    login_delete_obj_required = True
    login_read_obj_required = True

    read_obj_permission = True
    create_obj_permission = True
    update_obj_permission = True
    delete_obj_permission = True

    def has_login_read_obj_required(self):
        return self.login_required and self.login_read_obj_required

    def has_login_create_obj_required(self):
        return self.login_required and self.login_create_obj_required

    def has_login_update_obj_required(self):
        return self.login_required and self.login_update_obj_required

    def has_login_delete_obj_required(self):
        return self.login_required and self.login_delete_obj_required

    def has_create_obj_permission(self, obj=None, **kwargs):
        return ((not self.has_login_create_obj_required() or self.request.user.is_authenticated) and
                super(RESTObjectLoginMixin, self).has_create_obj_permission(obj=obj, **kwargs))

    def has_update_obj_permission(self, obj=None, **kwargs):
        return ((not self.has_login_update_obj_required() or self.request.user.is_authenticated) and
                super(RESTObjectLoginMixin, self).has_update_obj_permission(obj=obj, **kwargs))

    def has_delete_obj_permission(self, obj=None, **kwargs):
        return ((not self.has_login_delete_obj_required() or self.request.user.is_authenticated) and
                super(RESTObjectLoginMixin, self).has_delete_obj_permission(obj=obj, **kwargs))

    def has_read_obj_permission(self, obj=None, **kwargs):
        return ((not self.has_login_read_obj_required() or self.request.user.is_authenticated) and
                super(RESTObjectLoginMixin, self).has_read_obj_permission(obj=obj, **kwargs))


class RESTResourceMixin:

    register = False
    abstract = True

    @classproperty
    @classmethod
    def csrf_exempt(cls):
        return not cls.login_required

    def dispatch(self, request, *args, **kwargs):
        if hasattr(self, 'core'):
            self.core.init_rest_request(request)
        return super(RESTResourceMixin, self).dispatch(request, *args, **kwargs)

    @classmethod
    def __init_core__(cls, core, pattern):
        cls.core = core
        cls.pattern = pattern

    def _get_error_response(self, exception):
        """
        Trasform pyston exceptions to Is-core exceptions and raise it
        """
        response_exceptions = {
            MimerDataException: HTTPBadRequestResponseException,
            NotAllowedException: HTTPForbiddenResponseException,
            UnsupportedMediaTypeException: HTTPUnsupportedMediaTypeResponseException,
            Http404: Http404,
            ResourceNotFoundException: Http404,
            NotAllowedMethodException: HTTPMethodNotAllowedResponseException,
            DuplicateEntryException: HTTPDuplicateResponseException,
            ConflictException: HTTPDuplicateResponseException,
        }
        response_exception = response_exceptions.get(type(exception))
        if response_exception:
            raise response_exception
        return super(RESTResourceMixin, self)._get_error_response(exception)

    def _get_cors_allowed_headers(self):
        return super(RESTResourceMixin, self)._get_cors_allowed_headers() + (settings.AUTH_HEADER_NAME,)


class RESTModelCoreResourcePermissionsMixin(RESTObjectLoginMixin):

    pk_name = 'pk'

    def has_create_obj_permission(self, obj=None, **kwargs):
        return (super(RESTModelCoreResourcePermissionsMixin, self).has_create_obj_permission(obj=obj, **kwargs) and
                self.core.has_rest_create_permission(self.request, obj=obj, **kwargs))

    def has_update_obj_permission(self, obj=None, **kwargs):
        obj = obj or self._get_perm_obj_or_none()
        return (super(RESTModelCoreResourcePermissionsMixin, self).has_update_obj_permission(obj=obj, **kwargs) and
                self.core.has_rest_update_permission(self.request, obj=obj, **kwargs))

    def has_delete_obj_permission(self, obj=None, **kwargs):
        obj = obj or self._get_perm_obj_or_none()
        return (super(RESTModelCoreResourcePermissionsMixin, self).has_delete_obj_permission(obj=obj, **kwargs) and
                self.core.has_rest_delete_permission(self.request, obj=obj, **kwargs))

    def has_read_obj_permission(self, obj=None, **kwargs):
        obj = obj or self._get_perm_obj_or_none()
        return (super(RESTModelCoreResourcePermissionsMixin, self).has_read_obj_permission(obj=obj, **kwargs) and
                self.core.has_rest_read_permission(self.request, obj=obj, **kwargs))

    def _get_perm_obj_or_none(self, pk=None):
        pk = pk or self.kwargs.get(self.pk_name)
        if pk:
            return get_object_or_none(self.core.model, pk=pk)
        else:
            return None


class RESTModelCoreMixin(RESTModelCoreResourcePermissionsMixin):

    def _get_queryset(self):
        return self.core.get_queryset(self.request)

    def _get_pk(self):
        return self.kwargs.get(self.pk_name)

    def _get_obj_or_none(self, pk=None):
        if pk or self._get_pk():
            return get_object_or_none(self._get_queryset(), pk=(pk or self._get_pk()))
        else:
            return None

    def _get_obj_or_404(self, pk=None):
         obj = self._get_obj_or_none(pk)
         if not obj:
             raise Http404
         return obj


class RESTResource(RESTLoginMixin, RESTResourceMixin, BaseResource):
    pass


class EntryPointResource(RESTResource):

    login_required = False
    allowed_methods = ('get', 'head', 'options')

    def get(self):
        out = {}
        for pattern_name, pattern in patterns.items():
            if isinstance(pattern, RESTPattern):
                try:
                    url = pattern.get_url_string(self.request)
                    allowed_methods = pattern.get_allowed_methods(self.request, None)
                    if allowed_methods:
                        out[pattern_name] = {'url': url, 'methods': allowed_methods}
                except NoReverseMatch:
                    pass

        return out


class RESTModelResource(RESTModelCoreMixin, RESTResourceMixin, BaseModelResource):

    form_class = None
    field_labels = None
    abstract = True
    filters = {}
    default_fields_extension = None

    def _get_field_labels(self):
        return (
            self.field_labels if self.field_labels is not None else self.core.get_rest_form_field_labels(self.request)
        )

    def get_fields(self, obj=None):
        fields = super(DefaultRESTModelResource, self).get_fields(obj=obj)
        return self.core.get_rest_fields(self.request, obj=None) if fields is None else fields

    def get_detailed_fields(self, obj=None):
        detailed_fields = super(DefaultRESTModelResource, self).get_detailed_fields(obj=obj)
        return self.core.get_rest_detailed_fields(self.request, obj=obj) if detailed_fields is None else detailed_fields

    def get_general_fields(self, obj=None):
        general_fields = super(DefaultRESTModelResource, self).get_general_fields(obj=obj)
        return self.core.get_rest_general_fields(self.request, obj=obj) if general_fields is None else general_fields

    def get_guest_fields(self, obj=None):
        guest_fields = super(DefaultRESTModelResource, self).get_guest_fields(obj=obj)
        return self.core.get_rest_guest_fields(self.request, obj=obj) if guest_fields is None else guest_fields

    def get_extra_fields(self, obj=None):
        extra_fields = super(DefaultRESTModelResource, self).get_extra_fields(obj=obj)
        return self.core.get_rest_extra_fields(self.request) if extra_fields is None else extra_fields

    def get_default_fields(self, obj=None):
        default_fields = super(DefaultRESTModelResource, self).get_default_fields(obj=obj)
        return self.core.get_rest_default_fields(self.request, obj=None) if default_fields is None else default_fields

    def get_extra_filter_fields(self):
        extra_filter_fields = super(DefaultRESTModelResource, self).get_extra_filter_fields()
        return self.core.get_rest_extra_filter_fields(self.request) if extra_filter_fields is None else extra_filter_fields

    def get_filter_fields(self):
        filter_fields = super(DefaultRESTModelResource, self).get_filter_fields()
        return self.core.get_rest_filter_fields(self.request) if filter_fields is None else filter_fields

    def get_extra_order_fields(self):
        extra_order_fields = super(DefaultRESTModelResource, self).get_extra_order_fields()
        return self.core.get_rest_extra_order_fields(self.request) if extra_order_fields is None else extra_order_fields

    def get_order_fields(self):
        order_fields = super(DefaultRESTModelResource, self).get_order_fields()
        return self.core.get_rest_order_fields(self.request) if order_fields is None else order_fields

    def get_default_fields_rfs(self, obj=None):
        return super(RESTModelResource, self).get_default_fields_rfs(obj=obj).join(
            rfs(self.get_default_fields_extension(obj))
        )

    def get_default_fields_extension(self, obj=None):
        return (
            self.core.get_rest_default_fields_extension(self.request, obj=None)
            if self.default_fields_extension is None else self.default_fields_extension
        )

    def _rest_links(self, obj):
        rest_links = {}
        for pattern in self.core.rest_patterns.values():
            if pattern.send_in_rest:
                url = pattern.get_url_string(self.request, obj=obj)
                if url:
                    allowed_methods = pattern.get_allowed_methods(self.request, obj)
                    if allowed_methods:
                        rest_links[pattern.name] = {
                            'url': url,
                            'methods': [method.upper() for method in allowed_methods]
                        }
        return rest_links

    def _default_action(self, obj):
        return self.core.get_default_action(self.request, obj=obj)

    def _actions(self, obj):
        return self.core.get_list_actions(self.request, obj)

    def _class_names(self, obj):
        return self.core.get_rest_obj_class_names(self.request, obj)

    def get_queryset(self):
        return self.core.get_queryset(self.request)

    def _get_headers_queryset_context_mapping(self):
        mapping = super(RESTModelResource, self)._get_headers_queryset_context_mapping()
        mapping.update({
            'direction': ('HTTP_X_DIRECTION', '_direction'),
            'order': ('HTTP_X_ORDER', '_order')
        })
        return mapping

    def _preload_queryset(self, qs):
        return self.core.preload_queryset(self.request, qs)

    def _get_exclude(self, obj=None):
        return self.core.get_rest_form_exclude(self.request, obj)

    def _get_form_fields(self, obj=None):
        return self.core.get_rest_form_fields(self.request, obj)

    def _get_form_class(self, obj=None):
        return (
            self.form_class or
            (
                self.core.get_rest_form_edit_class(self.request, obj)
                if obj
                else self.core.get_rest_form_add_class(self.request, obj)
            )
        )

    def _get_form_initial(self, obj):
        return {'_request': self.request, '_user': self.request.user}

    def _pre_save_obj(self, obj, form, change):
        self.core.pre_save_model(self.request, obj, form, change)

    def _save_obj(self, obj, form, change):
        self.core.save_model(self.request, obj, form, change)

    def _post_save_obj(self, obj, form, change):
        self.core.post_save_model(self.request, obj, form, change)

    def _pre_delete_obj(self, obj):
        self.core.pre_delete_model(self.request, obj)

    def _delete_obj(self, obj):
        self.core.delete_model(self.request, obj)

    def _post_delete_obj(self, obj):
        self.core.post_delete_model(self.request, obj)

    def _generate_form_class(self, inst, exclude=[]):
        form_class = self._get_form_class(inst)
        exclude = list(self._get_exclude(inst)) + exclude
        fields = self._get_form_fields(inst)
        if hasattr(form_class, '_meta') and form_class._meta.exclude:
            exclude.extend(form_class._meta.exclude)
        return rest_modelform_factory(self.model, form=form_class, form_factory=smartmodelform_factory,
                                      auto_related_direct_fields=pyston_settings.AUTO_RELATED_DIRECT_FIELDS,
                                      auto_related_reverse_fields=pyston_settings.AUTO_RELATED_REVERSE_FIELDS,
                                      request=self.request, exclude=exclude, fields=fields,
                                      labels=self._get_field_labels())

    def put(self):
        # TODO: backward compatibility for bulk update should be used only patch
        return super(RESTModelResource, self).put() if self.kwargs.get(self.pk_name) else self.update_bulk()

    def patch(self):
        return super(RESTModelResource, self).patch() if self.kwargs.get(self.pk_name) else self.update_bulk()

    def _get_queryset(self):
        return self.core.get_queryset(self.request)

    @transaction.atomic
    def update_bulk(self):
        qs = self._filter_queryset(self._get_queryset())
        BULK_CHANGE_LIMIT = getattr(settings, 'BULK_CHANGE_LIMIT', 200)
        if qs.count() > BULK_CHANGE_LIMIT:
            return RESTErrorResponse(
                msg=ugettext('Only %s objects can be changed by one request').format(BULK_CHANGE_LIMIT),
                code=413)

        data = self.get_dict_data()
        objects, errors = zip(*(self._update_obj(obj, data) for obj in qs))
        compact_errors = tuple(err for err in errors if err)
        return RESTErrorsResponse(compact_errors) if len(compact_errors) > 0 else objects

    def _update_obj(self, obj, data):
        try:
            return (self._create_or_update(merge(data, {self.pk_field_name: obj.pk}), partial_update=True), None)
        except DataInvalidException as ex:
            return (None, self._format_message(obj, ex))
        except (ConflictException, NotAllowedException):
            raise
        except RESTException as ex:
            return (None, self._format_message(obj, ex))

    def _extract_message(self, ex):
        return '\n'.join([force_text(v) for v in ex.errors.values()]) if hasattr(ex, 'errors') else ex.message

    def _format_message(self, obj, ex):
        return {
            'id': obj.pk,
            'errors': {k: mark_safe(force_text(v)) for k, v in ex.errors.items()} if hasattr(ex, 'errors') else {},
            '_obj_name': force_text(obj),
        }


class UIRESTModelResource(RESTModelResource):

    def _web_links(self, obj):
        web_links = {}
        for pattern in self.core.web_link_patterns(self.request):
            if pattern.send_in_rest:
                url = pattern.get_url_string(self.request, obj=obj)
                if url and pattern.can_call_get(self.request, obj=obj):
                    web_links[pattern.name] = url
        return web_links

    def _class_names(self, obj):
        return self.core.get_rest_obj_class_names(self.request, obj)



from germanium.tools.rest import assert_valid_JSON_response

from is_core.config import settings
from is_core.utils import header_name_to_django


class RESTAuthMixin:

    def authorize(self, username, password):
        if settings.AUTH_USE_TOKENS:
            resp = self.post(settings.LOGIN_API_URL,
                             data={settings.USERNAME: username, settings.PASSWORD: password})
            assert_valid_JSON_response(resp, 'REST authorization fail: %s' % resp)
            self.default_headers[header_name_to_django(settings.AUTH_HEADER_NAME)] = (
                self.deserialize(resp).get('token')
            )
        else:
            super(RESTAuthMixin, self).authorize(username, password)


from collections import OrderedDict

from importlib import import_module

from django.conf import settings
from django.utils.encoding import force_text
from django.apps import apps


class App:

    def __init__(self):
        self.cores = []

    def add_core(self, core):
        if core not in self.cores:
            self.cores.append(core)


class CoresLoader:

    def __init__(self):
        self.apps = OrderedDict()

    def register_core(self, app_label, core):
        app = self.apps.get(app_label, App())
        app.add_core(core)
        self.apps[app_label] = app

    def _init_apps(self):
        import_module('is_core.main')

        for app in apps.get_app_configs():
            try:
                import_module('{}.cores'.format(app.name))
            except ImportError as ex:
                if force_text(ex) != 'No module named \'{}.cores\''.format(app.name):
                    raise ex

    def get_cores(self):
        self._init_apps()

        for app in self.apps.values():
            for core in app.cores:
                yield core

loader = CoresLoader()
register_core = loader.register_core
get_cores = loader.get_cores



from .commands import *  # NOQA
from .http_exceptions import *  # NOQA
from .models import *  # NOQA
from .rest_permissions import *  # NOQA
from .ui_ordering import *  # NOQA
from .ui_permissions import *  # NOQA



from django.core.exceptions import ImproperlyConfigured

from security.utils import get_throttling_validators
from security.throttling import UnsuccessfulLoginThrottlingValidator, SuccessfulLoginThrottlingValidator


try:
    LOGIN_THROTTLING_VALIDATORS = get_throttling_validators('is_core_login_validators')
except ImproperlyConfigured:
    LOGIN_THROTTLING_VALIDATORS = (
        UnsuccessfulLoginThrottlingValidator(60, 2),
        UnsuccessfulLoginThrottlingValidator(10 * 60, 10),
        SuccessfulLoginThrottlingValidator(60, 2),
        SuccessfulLoginThrottlingValidator(10 * 60, 10),
    )



from django.conf import settings as django_settings


DEFAULTS = {
    'AUTH_COOKIE_NAME': 'Authorization',
    'AUTH_COOKIE_AGE': 60 * 60 * 24 * 7 * 2,  # Age of cookie, in seconds (default: 2 weeks)
    'AUTH_COOKIE_HTTPONLY': False,
    'AUTH_COOKIE_SECURE': False,
    'AUTH_COOKIE_DOMAIN': None,
    'AUTH_HEADER_NAME': 'Authorization',
    'AUTH_DEFAULT_TOKEN_AGE': 60 * 60,  # Default token expiration time (default: 1 hour)
    'AUTH_MAX_TOKEN_AGE': 60 * 60 * 24 * 7 * 2,  # Max token expiration time (default: 2 weeks)
    'AUTH_COUNT_USER_PRESERVED_TOKENS': 20,
    'AUTH_USE_TOKENS': False,
    'AUTH_FORM_CLASS': lambda s: ('is_core.auth_token.forms.TokenAuthenticationForm' if s.AUTH_USE_TOKENS
                                  else 'django.contrib.auth.forms.AuthenticationForm'),
    'AUTH_RESOURCE_CLASS': 'is_core.auth_token.auth_resource.AuthResource',
    'AUTH_LOGIN_VIEW': lambda s: ('is_core.auth_token.auth_views.TokenLoginView' if s.AUTH_USE_TOKENS
                                  else 'class_based_auth_views.views.LoginView'),
    'AUTH_LOGOUT_VIEW': lambda s: ('is_core.auth_token.auth_views.TokenLogoutView' if s.AUTH_USE_TOKENS
                                   else 'is_core.generic_views.auth_views.LogoutView'),
    'AUTH_TAKEOVER_REDIRECT_URL': '/',
    'HOME_CORE': 'is_core.main.HomeUIISCore',
    'HOME_VIEW': 'is_core.generic_views.HomeView',
    'MENU_GENERATOR': 'is_core.menu.MenuGenerator',
    'USERNAME': 'username',
    'PASSWORD': 'password',
    'LOGIN_URL': django_settings.LOGIN_URL,
    'LOGOUT_URL': '/logout/',
    'LOGIN_API_URL': lambda s: '/api{}'.format(s.LOGIN_URL),
    'EXPORT_TYPES': '',
    'FOREIGN_KEY_MAX_SELECBOX_ENTRIES': 500,
    'LIST_PER_PAGE': 20,
    'AUTH_TOKEN_RENEWAL_EXEMPT_HEADER': 'X-Authorization-Renewal-Exempt',
    'AUTH_TOKEN_EXPIRATION_HEADER': 'X-Authorization-Expiration',
    'REST_DEFAULT_FIELDS_EXTENSION': ('_rest_links',),
    'RESPONSE_EXCEPTION_FACTORY': 'is_core.exceptions.response.ui_rest_response_exception_factory',
}


class Settings:

    def __getattr__(self, attr):
        if attr not in DEFAULTS:
            raise AttributeError('Invalid IS_CORE setting: "{}"'.format(attr))

        default = DEFAULTS[attr]
        return getattr(django_settings, 'IS_CORE_{}'.format(attr), default(self) if callable(default) else default)


settings = Settings()



import mimeparse

from django.http.response import HttpResponse, HttpResponseRedirect
from django.template.loader import render_to_string
from django.utils.encoding import force_text
from django.utils.translation import ugettext_lazy as _

from pyston.utils import set_rest_context_to_request
from pyston.utils.helpers import str_to_class
from pyston.converters import get_converter, get_supported_mime_types, get_converter_name_from_request
from pyston.resource import BaseResource
from pyston.conf import settings as pyston_settings

from is_core.config import settings
from is_core.utils import str_to_class


def ui_response_exception_factory(request, response_code, title, message, response_class=HttpResponse):
    if isinstance(message, (list, tuple)):
        message = ', '.join(message)

    return response_class(
        render_to_string(
            ('{}.html'.format(response_code), 'error.html'),
            {'response_code': response_code, 'title': title, 'message': message},
            request=request
        ),
        status=response_code
    )


def rest_response_exception_factory(request, response_code, title, message, response_class=HttpResponse):
    message = (
        ', '.join([force_text(val) for val in message])
        if isinstance(message, (list, tuple))
        else force_text(message)
    ) if message else force_text(title)
    set_rest_context_to_request(request, BaseResource.DEFAULT_REST_CONTEXT_MAPPING)
    converter_name = get_converter_name_from_request(request)
    converter = get_converter(converter_name)
    response = response_class(status=response_code, content_type=converter.content_type)

    rest_error_response = str_to_class(pyston_settings.ERROR_RESPONSE_CLASS)

    converter.encode_to_stream(response, rest_error_response(msg=message, code=response_code).result)
    return response


def ui_rest_response_exception_factory(request, response_code, title, message, response_class=HttpResponse):
    rest_mime_types = list(get_supported_mime_types())
    ui_mime_types = ['text/html', 'application/xhtml+xml']

    best_match = mimeparse.best_match(rest_mime_types + ui_mime_types, request.META.get('HTTP_ACCEPT', 'text/html'))

    if best_match in ui_mime_types:
        return ui_response_exception_factory(request, response_code, title, message, response_class)
    else:
        return rest_response_exception_factory(request, response_code, title, message, response_class)


def response_exception_factory(request, response_code, title, message, response_class=HttpResponse):
    return str_to_class(settings.RESPONSE_EXCEPTION_FACTORY)(request, response_code, title, message, response_class)


class ResponseException(Exception):

    def __init__(self, message=None):
        self.message = message

    def get_response(self, request):
        return response_exception_factory(request, self.status_code, self.title, self.message)


class HTTPRedirectResponseException(ResponseException):

    def __init__(self, url):
        self.url = url

    def get_response(self, request):
        return HttpResponseRedirect(self.url)


class HTTPBadRequestResponseException(ResponseException):

    title = _('Bad Request')
    status_code = 400


class HTTPUnauthorizedResponseException(ResponseException):

    title = _('Unauthorized')
    status_code = 401


class HTTPForbiddenResponseException(ResponseException):

    title = _('Forbidden')
    status_code = 403


class HTTPUnsupportedMediaTypeResponseException(ResponseException):

    title = _('Unsupported Media Type')
    status_code = 415


class HTTPMethodNotAllowedResponseException(ResponseException):

    title = _('Method Not Allowed')
    status_code = 405


class HTTPDuplicateResponseException(ResponseException):

    title = _('Conflict/Duplicate')
    status_code = 409


class HTTPServiceUnavailableException(ResponseException):

    title = _('Temporarily unavailable')
    status_code = 503
